/*
 * mpgedit application code
 *
 * Copyright (C) 2001-2003 Adam Bernstein
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA  02111-1307  USA.
 */


#ifndef lint
static char SccsId[] = "$Id: mp3_header.c,v 1.57 2003/09/11 06:14:11 number6 Exp $";
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <sys/stat.h>
#include "portability.h"
#include "header.h"
#include "mp3time.h"
#include "mpegindx.h"
#include "xing_header.h"
#include "mpegfio.h"
#include "mp3_header.h"
#include "mpegcurses.h"
#include "decoder.h"
#include "plugin.h"
#include "version.h"
#include "editif.h"
#include "playif.h"
#include "volumeif.h"
#include "mpegstat.h"


typedef struct _cmdflags
{
    editspec_t *edits;
    int      edits_cnt;
    char *out_filename;
    int  out_fileappend;
    int O_flag;
    int v_flag;
    int p_flag;
    int e_flag;
    int X_flag;
    char *X_args;
    int X_val;
    int c_flag;
    int s_flag;
    int has_xing;
    int L_flag;
    int l_flag;
    char *l_value;
} cmdflags;


typedef struct _mp3edit_play_cb_struct
{
    char  *filename;
    mpegfio_iocallbacks *ttyio;
    long  esec;
    long  emsec;
    long  tlast;
    long  frame_count;
    int   verbose;
    void *play_ctx;
    int   once;
} mp3edit_play_cb_struct;


static char *g_progname;

/* Leading plus forces POSIX processing of argument list */
#define OPTARGS "+hHf:vpe:X::o:csVl:L"

void usage(int longhelp)
{
    fprintf(stderr, "usage: %s %s\n", 
            g_progname, "-V | -h | -H | [-c [input_file [input_file2] ... ]]");
    fprintf(stderr, "       %s %s%s%s\n", g_progname,
         "[-L left[:right]] [-L] [-p] [-s] [-v[v[v]]] ",
         "[-X[0|1]] [-o [+]outfile_name] ",
         "[-e [begint[-[endt]]]] [-f] input_file");
    if (!longhelp) {
        exit(1);
    }

    if (longhelp == 2) {
        fprintf(stderr, 
"Where\n "
"  -c [input file [input_file2] ...]\n"
"      Curses edit mode.  Enter a full screen, interactive edit session\n"
"      to playback input files and set begin and end times for each\n"
"      edit.  This mode is the default action when no arguments\n"
"      are supplied.\n\n"
"  -estart[-[end]] \n"
"      Perform copy/paste operation on input files. The input MP3\n"
"      file may be encoded as either CBR or VBR.  For quick and accurate\n"
"      random access into VBR encoded files, an external time index file\n"
"      is created.  The index file uses the basename of the input file,\n"
"      with the .idx extension.\n"
"      \n"
"      The output files created by this option are located in the\n"
"      same directory as the input MP3 file.  When the -o option is\n"
"      not used, the copy/paste data is written to a file\n"
"      with the same basename as the input file, with the '_nn.mp3'\n"
"      extension, where nn is an integer value. When a file with\n"
"      basename_nn.mp3 already exists, the integer extension is incremented\n"
"      until the filename is unique.  The input MP3 file is never\n"
"      overwritten by this option.\n"
"      \n"
"      When just a begin time is specified, the copy/paste operation is\n"
"      performed from that begin time through the end of file. When just\n"
"      an end time is specified, the copy/paste operation is performed\n"
"      from the beginning of the file through the specified end time.\n"
"      When both begin and end times are specified, the portion of the\n"
"      file bracketed by these times is copy/pasted.  Begin and end times\n"
"      are specified in seconds, or MM:SS.mmm, where MM are minutes, SS\n"
"      are seconds, and mmm are milliseconds.  Although any millisecond\n"
"      value may be specified, it will always be rounded off to the next\n"
"      frame boundary, approximately 26ms.\n"
"      \n"
"      Multiple -e options are allowed. Each -e represents a separate edit\n"
"      that will be performed on the named input file that follows the last\n"
"      -f option.\n\n"
"  -f  input file name\n"
"      Multiple -f options are allowed.  At least one -e option must be\n"
"      preceed each -f option.\n\n"
"  -o  output file\n"
"      Prefixing a + to filename causes output from edit to be appended\n"
"      to the output file.  XING headers are automatically fixed up when\n"
"      using this option.\n\n"
"  -h  command usage summary\n\n"
"  -H  verbose usage information; this help screen\n\n"
"  -l  left[:right]\n"
"      Set the current playback device volume level to a value between 0 and\n"
"      100.  Specifying only a left value sets both left and right channels\n"
"      to that value.\n\n"
"  -L  Display the current playback device volume level.\n\n"
"  -p  play MPEG file using external player\n\n"
"  -s  silent operation\n"
"      All of the normal informational messages are suppressed during\n"
"      indexing and editing.\n\n"
"  -v  verbose output; adding more -v increases verbosity.\n"
"      Up to -vvv is allowed. -v displays the frame count and file\n"
"      offset at one second increments. -vv adds to -v this information\n"
"      for each frame: time offset, bitrate, frame size, frame count,\n"
"      frame position.  -vvv adds to -vv and displays a detailed breakdown\n"
"      of each MPEG frame header.\n\n"
"  -V  display mpgedit version and build number\n\n"
"  -X[0|1]\n"
"      Manipulate XING header. 0 omit XING header prefix, allowing output\n"
"      to be appended to the end of a previous edit. 1 fixes up XING header\n"
"      after catenating edits. This option is largely obsolete because of '+'\n"
"      option to -o\n"
);
    }
    else {
        fprintf(stderr, "\n    For detailed mpgedit usage, use -H\n");
    }
    exit(1);
}




/*
 * Support function for mp3edit_find_free_slot().
 */
static int mp3edit_file_exists(char *base, const char *ext, char *offset, int n)
{
    sprintf(offset, "_%d.%s", n, ext);
    return access(base, F_OK) == 0;
}


static char *mp3edit_make_cached_slot_name(const char *base, const char *ext)
{
    char *cp;
    char *tp;

    cp = (char *) malloc(strlen(base)+strlen(ext)+5);
    if (!cp) {
        return NULL;
    }
    
    /* 
     * Transform path: path/file -> path/.file_#
     *                      file -> .file_#
     */
    cp[0] = '\0';
    tp = strrchr(base, '/');
    if (tp) {
        strncat(cp, base, tp-base+1);
        strcat(cp, ".");
        strcat(cp, tp+1);
    }
    else {
        strcpy(cp, ".");
        strcat(cp, base);
    }

    strcat(cp, ".");
    strcat(cp, ext);
    strcat(cp, "_#");
    return cp;
}


static int mp3edit_find_cached_free_slot(const char *name_cursor,
                                         const char *base,
                                         const char *ext,
                                         int *sequence)
{
    FILE *fp = NULL;
    char buf[1024];
    char *cp;
    char *endcp;
    int  n;
    int  rsts = 0;

    fp = fopen(name_cursor, "r+");
    if (!fp) {
        rsts = 1;
        goto clean_exit;
    }
    cp = fgets(buf, sizeof(buf)-1, fp);
    if (!cp) {
        rsts = 1;
        goto clean_exit;
    }

    /* Stomp trailing '\n' */
    cp[strlen(cp)-1] = '\0';

    /* Check that cached file exists.  If not, out of sync */
    if (access(cp, F_OK) == -1) {
        unlink(name_cursor);
        rsts = 1;
        goto clean_exit;
    }

    cp = strrchr(buf, '.');
    if (cp) {
        cp = strrchr(buf, '_');
    }
    if (!cp) {
        rsts = 1;
        goto clean_exit;
    }

    /* Hopefully, cp is now positioned at _NNN.mp3 in cached name */
    cp++;
    n = strtol(cp, &endcp, 10);
    if (*endcp != '.' || strcmp(endcp+1, ext)) {
        rsts = 1;
        goto clean_exit;
    }
    n++;
    sprintf(cp, "%d.%s", n, ext);

    
    /*
     * Check that next sequence numberef file exists.  If it does, out of sync
     */
    if (access(buf, F_OK) == 0) {
        unlink(name_cursor);
    }
    else {
        fseek(fp, 0, SEEK_SET);
        fprintf(fp, "%s\n", buf);

        /* Return next sequence number */
        *sequence = n;
    }

clean_exit:
    if (fp) {
        fclose(fp);
    }

    
    return rsts;
}


static int mp3edit_write_cached_slot_name(const char *name_cursor, char *base)
{
    int rsts = 0;
    FILE *fp;

    fp = fopen(name_cursor, "w");
    if (!fp) {
        rsts = 1;
    }
    else {
        fprintf(fp, "%s\n", base);
        fclose(fp);
    }
    return rsts;
}


/*
 * Find the next available output file.   Given a file with the name
 * name.mp3, this function builds the name name_NNN.mp3, where NNN
 * is an integer number.  This function finds the next unused file
 * name in the NNN sequence. 
 *
 * The search for the next available file is implemented as a binary
 * search.  The search starts by finding the first unused filename, by
 * incrementing NNN by powers of 2.  Once this high number is found,
 * then the binary search begins.
 *
 * This function modifies the space pointed at by base.  This
 * function is an internal helper routine for build_edit_filename();
 *
 * Support function for build_edit_filename().
 */
static char *mp3edit_find_free_slot(char *in_base, const char *ext)
{
    int err  = 0;
    int low  = 1;
    int high = 1;
    int test;
    int rsts;
    char *offset;
    char *base       = NULL;
    char *cache_file = NULL;

    /*
     * Convert /path/prefix.mp3 -> /path/prefix_N.mp3.
     * New size is length of path + prefix + "_" (1) + max integer
     * size (11) + ".ext" + null byte (1).
     */
    base = (char *) calloc(1, strlen(in_base) + 1 + 11 + 1 + strlen(ext) + 1);
    if (!base) {
        return NULL;
    }
    strcpy(base, in_base);
    offset = base + strlen(base);

    /*
     * Search for first available unused slot
     */
    cache_file = mp3edit_make_cached_slot_name(base, ext);
    if (!cache_file) {
        err = 1;
        goto clean_exit;
    }

    rsts = mp3edit_find_cached_free_slot(cache_file, base, ext, &test);
    if (rsts == 0) {
       /* Format the return file name using the cached slot number */
       sprintf(offset, "_%d.%s", test, ext);
       goto clean_exit;
    }

    rsts = mp3edit_file_exists(base, ext, offset, high);
    while (rsts) {
        low = high;
        high *= 2;
        rsts = mp3edit_file_exists(base, ext, offset, high);
    }

    /* found slot if first numbered file does not exit */
    if (low == high) {
        goto clean_exit;
    }

    /* Perform binary search for next available slot */
    do {
        test = low + ((high - low) / 2);
        rsts = mp3edit_file_exists(base, ext, offset, test);
        if (rsts) {
            low = test;
        }
        else {
            high = test;
        }
    } while ((low+1) != high);

    /* Format the file name using the high value */
    mp3edit_file_exists(base, ext, offset, high);
    mp3edit_write_cached_slot_name(cache_file, base);
    free(cache_file);

clean_exit:
    if (err) {
        if (base) {
            free(base);
        }
    }
    return base;
}


/*
 * Given an input path to an mp3 file, return the path to the corresponding
 * edit output file.  One rule applies: when the directory containing the file
 * is read-only, the returned output file is in the local directory.
 * This function builds a unique output file name, by adding the suffix
 * _NNN between the filebase and the file extension.  Given /path/file.mp3,
 * the output file name /path/file_1.mp3 is created.  Should that file exist,
 * the integer suffix is incremented until the file does not exist.
 *
 * API function called by main().
 */
static char *build_edit_filename(const char *infile_name, const char *suffix)
{
    char *edit_filename = NULL;
    char *dir;
    char *filebase = NULL;
    char *ext      = NULL;
    int  len;
    char *ret_filename = NULL;

    dir = parse_pathname(infile_name, &filebase, &ext);
    if (!dir) {
        goto clean_exit;
    }

    /*
     * Is the directory read-only?
     */
    if (access(dir, W_OK) == -1) {
        free(dir);
        dir = strdup(".");
        if (!dir) {
            goto clean_exit;
        }
        /*
         * Is the directory still read-only? If so, fatal error, because
         * cannot write to the local directory.
         */
        if (access(dir, W_OK) == -1) {
            goto clean_exit;
        }
    }

    /*
     * Convert /path/prefix.mp3 -> /path/prefix_N.mp3.
     * New size is length of path + prefix + "_" (1) + max integer
     * size (11) + ".ext" + null byte (1).
     */
    len = strlen(dir) + 1 + strlen(filebase) + 1 + 11 + 1 + strlen(ext) + 1;
    if (suffix) {
        len += strlen(suffix);
    }
    edit_filename = (char *) calloc(1, len);
    if (!edit_filename) {
        fprintf(stdout, "failed to alloc edit filename\n");
        goto clean_exit;
    }
    strcpy(edit_filename, dir);
    if (strlen(dir) != 2 || dir[1] != ':') {
        strcat(edit_filename, "/");
    }
    strcat(edit_filename, filebase);
    if (suffix) {
        strcat(edit_filename, suffix);
    }
    if (!ext || !*ext) {
        if (ext) {
            free(ext);
        }
        ext = strdup("mp3");
        if (!ext) {
            goto clean_exit;
        }
    }
    ret_filename = mp3edit_find_free_slot(edit_filename, ext);
    free(edit_filename);

clean_exit:
    if (!dir || !filebase || !ext) {
        if (ret_filename) {
            free(ret_filename);
            ret_filename = NULL;
        }
    }

    if (dir) {
        free(dir);
    }
    if (filebase) {
        free(filebase);
    }
    if (ext) {
        free(ext);
    }
    return ret_filename;
}


static void myprintf(void *ctx, int y, int x, const char *fmt, ...)
{
    va_list ap;
    va_start(ap, fmt);
    vprintf(fmt, ap);
    fflush(stdout);
    va_end(ap);
}


/*
 * Not an exhaustive validation of the provided time specification,
 * but a consistency check.
 */
static int validate_edit_timespec(char *str)
{
    /* Null pointer or null string not valid */
    if (!str || !*str) {
        return 0;
    }

    /* 12 || :34 check */
    if (isdigit(str[0]) || str[0] == ':') {
        return 1;
    }

    /*
     * Must look at character following - to prove the value is not
     * the next option on the command line; -, -1:34 and -34 are valid.
     */
    if (str[0] == '-') {
        if (str[1] != '\0' && str[1] != ':' && !isdigit(str[1])) {
            return 0;
        }
        return 1;
    }

    return 0;
}


static void mpgedit_version(void)
{
    printf("\n%s version %s %s\n",
           MPGEDIT_PRODUCT,
           MPGEDIT_VERSION,
           MPGEDIT_URL);
    printf("Build:          %s (%s)\n", MPGEDIT_BUILDNUM, MPGEDIT_BUILDOS);
    printf("Report bugs to: %s\n", MPGEDIT_MAILTO);
    exit(1);
}


static char *mp3edit_alloc_infile(char *name)
{
    char *filename;
    char *cp;

    filename = strdup(name);
    if (!filename) {
        return NULL;
    }

    /*
     * Make all \ directory separators a /, as much of the subdirectory
     * processing code assumes a /, and Win32 filesystem functions will
     * accept a /.
     */
    for (cp = filename; *cp; cp++) {
        if (*cp == '\\') {
            *cp = '/';
        }
    }
    return filename;
}



static int index_callback(void *data, long sec, long msec,
                          long frame, char *indx_filename)
{
    mpegfio_iocallbacks *ttyio = (mpegfio_iocallbacks *) data;

    if (!data || !ttyio->printf) {
        return 1;
    }
    if (!ttyio->cb_called) {
        ttyio->printf(ttyio->ctx, 2, 0, 
                      "Creating index file %s\n", indx_filename);
        ttyio->cb_called = 1;
    }
    ttyio->printf(ttyio->ctx, 2, 0,
                  "\rElapsed time: %4d:%02d| %5d.%03ds|Frame: %-7d\r",
                  sec/60,
                  sec%60,
                  sec,
                  msec,
                  frame);
    return 1;
}


#if 1
/*
 * adam/TBD: Perhaps this function should be converted to a formatting
 * 2str function that lives in mpgeditlib.  That would allow more of
 * the internal library functions, like mpeg_time_gettime(), to not
 * be exported from the shared library.  This conversion is inevitable
 * when this verbose functionality is added to xmpgedit.
 */
#endif
static void mp3header_print_verbose(void *edit_ctx,
                                    int  verbose,
                                    int  *once,
                                    long *tprev)
{
    mpeg_header_data *stats_header;
    long             tnow = 0;
    mpeg_file_stats  *stats_ptr;
    mpeg_time        *file_time;
    long             sec;
    long             usec;
    long             frame_start;
    char             *buf;

    buf = malloc(1024);
    if (!buf) {
        return;
    }
    if (!(*once)) {
        *once = 1;
        if (verbose > 0 &&
            mpgedit_edit_has_xing(edit_ctx))
        {
            printf("Found xing header\n");
            xingheader2str(mpgedit_edit_xing_header(edit_ctx), buf);
            printf(buf);
        }
    }
    stats_header = mpgedit_edit_frame_header(edit_ctx);
    tnow         = mpgedit_edit_sec(edit_ctx);
    stats_ptr    = mpgedit_edit_stats_ctx(edit_ctx);
    file_time    = mpgedit_edit_total_length_time(edit_ctx);
    mpeg_time_gettime(file_time, &sec, &usec);

    if (verbose > 0) {
        if (verbose > 2) {
            printf("MPEG header ");
        }

        if (verbose > 1) {
            frame_start = stats_ptr->file_size - stats_header->frame_size;
            printf("t=%d.%-.3ds  br=%-3d  sz=%-4d "
                   "fr=%-6ld pos=%-10ld pos=0x%lx\n",
                   (int) sec, (int) (usec/1000),
                   stats_header->bit_rate,
                   stats_header->frame_size,
                   stats_ptr->num_frames,
                   frame_start, frame_start);
        }

        if (verbose > 2) {
            mpeg_header_values2str(stats_header, buf);
            printf(buf);
        }
        if (*tprev != tnow) {
            *tprev = tnow;
            printf("  *** t=%lds fr=%ld pos=%ld (0x%lx)\n",
                   tnow, stats_ptr->num_frames,
                   stats_ptr->file_size, stats_ptr->file_size);
        }
    }
    else if (verbose != -1) {
        if (*tprev != tnow) {
            *tprev = tnow;
            printf("\rElapsed time: %4ld:%02ld| %5ld.%03lds|Frame: %-7ld\r",
                   tnow/60,
                   tnow%60,
                   tnow,
                   usec/1000,
                   stats_ptr->num_frames);
            fflush(stdout);
        }
    }
    free(buf);
}


static int play_callback_func(void *data, long sec, long msec)
{
    int go = 1;

    mp3edit_play_cb_struct *ctx = (mp3edit_play_cb_struct *) data;

    if (sec > ctx->esec || (sec == ctx->esec && msec >= ctx->emsec)) {
        go = 0;
    }
    if (go) {
        mp3header_print_verbose(
            ctx->play_ctx,
            ctx->verbose,
            &ctx->once,
            &ctx->tlast);
    }

    return go;
}


/*
 * Create index file for named mp3 file.  When not NULL, the function
 * pointer 'callback' is called for each second of file processed, passed
 * the user-provided 'callback_data', and the current file second/millisecond
 * offset, the current frame number and the name of the index file.
 * When not NULL, after the file is indexed, the file file statistics,
 * and total length are returned.
 */
int mp3header_index_create(char *filename, 
                           int (*callback)(void *,   /* callback_data */
                                           long,     /* sec */
                                           long,     /* msec */
                                           long,     /* frame_count */
                                           char *),  /* index filename */
                           void *callback_data,
                           mpeg_file_stats *rstats,
                           mpeg_time *rtime)
{
    char                *indx_filename = NULL;
    void                *ctx;
    long                sec;
    long                msec;
    long                tprev = -1;
    long                frame_count;
    int                 cbrsts;
    int                 eof;
    int                 rsts = 0;

    indx_filename = build_index_filename(filename);
    if (!indx_filename) {
        return 1;
    }

    ctx    = mpgedit_edit_index_init(filename);
    eof    = mpgedit_edit_index(ctx);
    cbrsts = 1;
    while (cbrsts) {
        sec         = mpgedit_edit_index_sec(ctx);
        msec        = mpgedit_edit_index_msec(ctx);
        frame_count = mpgedit_edit_index_frames(ctx);
        if (callback && sec > tprev) {
            tprev = sec;
            cbrsts = callback(callback_data, sec,
                              msec, frame_count, indx_filename);
        }

        if (eof) {
            cbrsts = 0;
        }

        if (cbrsts) {
            eof = mpgedit_edit_index(ctx);
        }
    }

    if (ctx) {
        if (rstats) {
            *rstats = *mpgedit_edit_stats_ctx(ctx);
        }
        if (rtime) {
            *rtime = *mpgedit_edit_total_length_time(ctx);
        }
        mpgedit_edit_index_free(ctx);
    }
    else {
        rsts = 1;
    }

    free(indx_filename);
    return rsts;
}


static void play_files(editspec_t *next_editspec, cmdflags *argvflags)
{
    int indx;
    void                   *play_ctx = NULL;
    mp3edit_play_cb_struct play_cb_data;
    int                    go;
    mpeg_file_stats        *stats_ptr;
    char                   *str;
    long                   sec;
    long                   usec;

    str = malloc(1024);
    if (!str) {
        return;
    }
    for (indx=0; indx<mpgedit_editspec_get_length(next_editspec); indx++) {
        /*
         * play files
         */
        play_ctx = mpgedit_play_init(mpgedit_editspec_get_file(
                                     next_editspec, indx));

        mpeg_time_gettime(mpgedit_editspec_get_stime(next_editspec, indx),
                          &sec, &usec);
        mpgedit_play_seek_time(play_ctx, 
                               sec, usec/1000);
        memset(&play_cb_data, 0, sizeof(play_cb_data));
        mpeg_time_gettime(mpgedit_editspec_get_etime(next_editspec, indx),
                          &sec, &usec);
        play_cb_data.esec      = sec;
        play_cb_data.emsec     = usec/1000;
        play_cb_data.verbose   = argvflags->s_flag ? -1 :
                                 argvflags->v_flag;
        play_cb_data.play_ctx  = play_ctx;
        
        mpgedit_play_set_status_callback(play_ctx,
                                         play_callback_func,
                                         &play_cb_data);
        do {
            go = mpgedit_play_frame(play_ctx);
        } while (go);
        stats_ptr = mpgedit_edit_stats_ctx(play_ctx);
        mpeg_file_stats2str(stats_ptr,
                            mpgedit_edit_total_length_time(play_ctx),
                            str);
        printf(str);
          
        mpgedit_play_close(play_ctx);
    }
    free(str);
}


int main(int argc, char *argv[])
{
    FILE             *editfp = NULL;
    int              ch;
    mpeg_time        mpegtval;
    char             *edit_filename = NULL;
    editspec_t       *curs_edits;
    int              curs_edits_cnt;
    int              status;
    mpegfio_iocallbacks ttyio;
    cmdflags         argvflags;
    char             *errstr;

    int              editing;
    void             *edit_ctx;
    void             *play_ctx = NULL;
    int              rsts = 0;
    int              edit_flags = 0;
    int              edit_indx;
    editspec_t       *next_editspec;
    int              next_editspec_len;
    int              outfile_indx;
    char             *edit_outfilename = NULL;
    char             *cp;
    long             tprev = 0;
    mpeg_file_stats  *stats_ptr;
    mpeg_file_stats  indx_stats;
    int              once = 0;
    int              indx;
    int              no_edits_msg = 0;
    int              vol_left;
    int              vol_right;
    void             *volctx;
    int              edits_len;
    char             *str;

    g_progname = (g_progname = strrchr(argv[0], '/')) ?
                     g_progname+1 : argv[0];

    memset(&argvflags, 0, sizeof(argvflags));
    memset(&ttyio, 0, sizeof(ttyio));

    /* 
     * More memory is allocated than needed here, but this saves
     * having to pre-count the total number of -e options specified.
     */
    argvflags.edits = mpgedit_editspec_init();
    if (!argvflags.edits) {
        perror("mpgedit main: mpgedit_editspec_init failed");
        return 1;
    }
    opterr = 0;
    while ((ch = getopt(argc, argv, OPTARGS)) != -1) {
        switch (ch) {
          case 'h':
            usage(1);
            break;

          case 'H':
            usage(2);
            break;

          case 'f':
            mpgedit_editspec_append(argvflags.edits,
                                    mp3edit_alloc_infile(optarg), NULL);
            break;

          case 'o':
            if (argvflags.out_filename) {
                fprintf(stderr, "Error: can only specify -o once\n");
                usage(0);
            }
            argvflags.out_filename = strdup(optarg);
            break;

          case 'p':
            argvflags.p_flag = 1;
            break;

          case 'v':
            argvflags.v_flag++;
            break;

          case 'e':
            argvflags.e_flag = 1;
            if (optarg) {
                if (!validate_edit_timespec(optarg)) {
                    fprintf(stderr, "%s '%s' %s\n",
                            "Error: -e argument",
                            optarg,
                            "is not a valid time specification");
                    usage(0);
                }
                mpgedit_editspec_append(argvflags.edits, NULL, optarg);
            }
            break;

          case 'X':
            /*
             * XING header processing. -X is same as -X0.  -X0 prevents
             * addition of XING header to edited cut. -X1 performs analysis of
             * edited cut and fixes up header. -X2 restores Xing header to
             * file, then updates it, ala -X1 analysis.
             */
            argvflags.X_flag = 1;
            if (optarg) {
                argvflags.X_args = strdup(optarg);
                argvflags.X_val  = atoi(optarg);
            }

            if (argvflags.X_val == 2) {
                /*
                 * X2 logic
                 */
                argvflags.e_flag = 1;
                mpgedit_editspec_append(
                    argvflags.edits,
                    mp3edit_alloc_infile(XING_HEADER_FILE),
                    "-");
                xingheader_file_make(XING_HEADER_FILE);
            }
            break;

          case 's':
            argvflags.s_flag = 1;
            break;

          case 'c':
            argvflags.c_flag = 1;
            break;

          case 'V':
            mpgedit_version();
            break;

        case 'l':
          argvflags.l_flag = 1;
          argvflags.l_value = strdup(optarg);
          break;

        case 'L':
          argvflags.L_flag = 1;
          break;

          case '?':
            fprintf(stderr, "%s: '-%c' unknown option or missing argument\n", 
                    g_progname, optopt);
            usage(0);
            break;

          default:
            fprintf(stderr, "%s: error parsing command line\n", g_progname);
            usage(0);
            break;
        }
    }
    if (argc == 1) {
        argvflags.c_flag = 1;
    }


    /*
     * It is convenient to treat the first argument following all of the
     * command line options as the input file name.  
     */
    while (optind < argc && argv[optind]) {
        /* If any option is found, that is a fatal error! */
        if (argv[optind][0] == '-') {
            fprintf(stderr, "Unexpected option found after filenames '%s'\n",
                    argv[optind]);
            fprintf(stderr, "Try using '-f' before input file names\n");
            return 1;
        }

        mpgedit_editspec_append(argvflags.edits,
                                mp3edit_alloc_infile(argv[optind]), "");
        optind++;
    }

    if (argvflags.p_flag || argvflags.c_flag) {
        /*
         * Open/close decoder to reveal any decoder open errors.  Also do this
         * when entering curses mode, since it is highly likely the user will
         * want to use the playback feature while in curses mode.
         */
        play_ctx = mpgedit_play_init(NULL);
        if (!play_ctx) {
            fprintf(stderr, "Unable to open playback plugin\n");
            return 1;
        }
        mpgedit_play_close(play_ctx);
    }

    if (argvflags.L_flag || argvflags.l_flag || argvflags.c_flag) {
        volctx = mpgedit_volume_open();
        if (volctx) {
            if (argvflags.l_flag) {
                vol_left = vol_right = -1;
                vol_left = strtol(argvflags.l_value, &cp, 10);
                if (cp && *cp == ':') {
                    cp++;
                    vol_right = strtol(cp, NULL, 10);
                }
                else {
                    vol_right = vol_left;
                }
                if (vol_left < 0 || vol_left > 100) {
                    vol_left = 0;
                }
                if (vol_right < 0 || vol_right > 100) {
                    vol_right = 0;
                }
                mpgedit_volume_set(volctx, vol_left, vol_right);
            }
            if (argvflags.L_flag || argvflags.l_flag) {
                mpgedit_volume_get(volctx, &vol_left, &vol_right);
                printf("Playback device volume level %d:%d\n",
                       vol_left, vol_right);
            }
        }
    }

    if (argvflags.c_flag) {
        status = curs_play(argvflags.edits,
                           mpgedit_editspec_get_length(argvflags.edits),
                           &ttyio, volctx,
                           &curs_edits, &curs_edits_cnt,
                           &argvflags.O_flag,
                           &argvflags.out_filename);
        if (status) {
            return 1;
        }
        mpgedit_editspec_free(argvflags.edits);
        argvflags.edits  = curs_edits;
        argvflags.e_flag = 1;
    }

    /* Silent operation */
    if (argvflags.s_flag) {
        argvflags.v_flag = 0;
        ttyio.printf     = NULL;
        ttyio.getch      = NULL;
    }
    else {
        ttyio.printf = myprintf;
        ttyio.getch  = NULL;
    }


    if (argvflags.out_filename &&
        (argvflags.p_flag || (argvflags.X_flag && argvflags.X_val < 2)))
    {
        fprintf(stderr, "%s: -o not allowed with -p or -X\n", g_progname);
        return 1;
    }

    /* -o with no -e not allowed */
    if ((!argvflags.X_flag || (argvflags.X_flag && argvflags.X_val < 2)) &&
        !argvflags.e_flag && argvflags.out_filename)
    {
        fprintf(stderr, "%s: -o must be used with -e\n", g_progname);
        return 1;
    }

    edits_len = mpgedit_editspec_get_length(argvflags.edits);
    if (!edits_len) {
        fprintf(stderr, "%s: no file name specified\n", g_progname);
        return 1;
    }

    if (argvflags.p_flag && argvflags.X_flag) {
        fprintf(stderr, "%s: both -p and -X not allowed\n", g_progname);
        return 1;
    }

    str = malloc(1024);
    if (!str) {
        fprintf(stderr, "failed allocating print string buffer\n");
        return 1;
    }

    /*
     * curs_play() has already done this initialization and
     * validation, so don't duplicate it here.
     */
    if (argvflags.c_flag == 0) {
        /*
         * Process -e and -f options.
         */
        if (!mpgedit_edit_times_init(argvflags.edits)) {
            fprintf(stderr, "%s: -f filename never specified\n", g_progname);
            return 1;
        }

        status = 0;
        if (edits_len > 0) {
            errstr = NULL;
            for (indx=0; !errstr && indx<edits_len; indx++) {
                if (!argvflags.p_flag && !argvflags.e_flag &&
                    !argvflags.v_flag && !no_edits_msg) 
                {
                    printf("No edits specified; only reading input data\n\n");
                    no_edits_msg = 1;
                }

                ttyio.cb_called = 0; 
                mpeg_time_clear(&mpegtval);
                rsts = mp3header_index_create(
                           mpgedit_editspec_get_file(
                               argvflags.edits, indx),
                           index_callback,
                           &ttyio, &indx_stats, &mpegtval);
                /*
                 * Set indexed flag here to indicate the file statistics
                 * have been printed here.  Prevents reading file a 
                 * second time.
                 */
                mpgedit_editspec_set_indexed(
                    argvflags.edits, indx, ttyio.cb_called);


                if (!argvflags.p_flag && !argvflags.e_flag &&
                    ttyio.cb_called && rsts == 0) 
                {
                    printf("\ndone.\n");
                    if (!argvflags.v_flag) {
                        mpeg_file_stats2str(&indx_stats, &mpegtval, str);
                        printf(str);
                    }
                }
                errstr = mpgedit_edit_validate_times(argvflags.edits, indx, 1);
            }
        }
        if (errstr) {
            fprintf(stderr, "Error: %s\n", errstr);
            free(errstr);
            return 1;
        }
    }

    if (argvflags.out_filename) {
        /*
         * Populate edit_filename and out_fileappend when out_filename
         * is specified
         */
        if (argvflags.out_filename[0] == '+') {
            edit_filename = strdup(&argvflags.out_filename[1]);
            argvflags.out_fileappend = 1;
        }
        else {
            edit_filename = strdup(argvflags.out_filename);
            argvflags.out_fileappend = 0;
        }
        rsts = validate_outfile(argvflags.out_filename,
                                argvflags.out_fileappend, &errstr);
        if (rsts) {
            fprintf(stderr, "%s\n", errstr);
            return rsts;
        }
    }

    /*
     * When only an input file is specified, add -v action.
     */
    if (!argvflags.p_flag && !argvflags.e_flag) {
        if (!argvflags.v_flag && !no_edits_msg) {
            printf("No edits specified; only reading input data\n\n");
        }
        edit_flags |= MPGEDIT_FLAGS_NO_EDITS;
    }

    edit_indx    = 0;
    edit_ctx     = NULL;
    outfile_indx = 1;

    /* When very verbose, print output from every frame */
    if (argvflags.v_flag > 1) {
        edit_flags |= MPGEDIT_FLAGS_EACH_FRAME;
    }

    if (edit_filename) {
        edit_outfilename = strdup(edit_filename);
    }
    else {
        if (mpgedit_editspec_get_file(argvflags.edits, 0)) {
            edit_filename =
                strdup(mpgedit_editspec_get_file(argvflags.edits, 0));
            cp = strrchr(edit_filename, '.');
            if (cp) {
                *cp = '\0';
            }
        }
    }

    if (argvflags.X_flag) {
        if (argvflags.X_val == 0) {
            edit_flags |= MPGEDIT_FLAGS_XING_OMIT;
        }
        else if (argvflags.X_val == 1) {
            edit_flags |= MPGEDIT_FLAGS_XING_FIXUP;
        }
        else if (argvflags.X_val == 2) {
            if (edits_len != 2) {
                fprintf(stderr,
                        "%s: Must specify exactly one input "
                        "file with -X2 option\n", g_progname);
                return 1;
            }
            /*
             * For -X2 option, create unique output file name based on input
             * file name to save output with restored Xing header into.
             */
            if (!argvflags.out_filename) {
                edit_outfilename = build_edit_filename(
                                       mpgedit_editspec_get_file(
                                           argvflags.edits, 1), "_newxing");
                argvflags.out_filename = strdup(edit_outfilename);
            }
        }
    }

    do {
#if 1
        next_editspec = mpgedit_editspec_get_edit(
                            argvflags.edits,
                            edit_indx,
                            edits_len,
                            0, &next_editspec_len);
#else
        next_editspec = mpgedit_editspec_get_edit(
                            &argvflags.edits[edit_indx],
                            argvflags.edits_cnt - edit_indx,
                            0, &next_editspec_len);
#endif
        if (!next_editspec) {
/* adam/TBD: allocation error */
        }

        if (!argvflags.out_filename) {
/*
 * adam/TBD: This is where you will add special logic for new
 * option that disables the automatic sequential numbering of the
 * output files.
 */
#if 1
            edit_outfilename = build_edit_filename(edit_filename, NULL);
#else
            edit_outfilename = mpgedit_edit_make_output_filename(
                                   ".", edit_filename,
                                   "mp3", &outfile_indx);
#endif
        }

        if (argvflags.p_flag) {
            play_files(next_editspec, &argvflags);
            tprev = 0; /* adam/TBD: what is tprev used for ?? */
        }
        /*
         * Edits or verbose file statistics have been specified, or the file
         * statistics have not already been printed by the indexing operation
         * when no file operation has been specified.
         */
        else if (argvflags.e_flag || argvflags.v_flag || 
                 (next_editspec_len == 1 && 
                  mpgedit_editspec_get_indexed(next_editspec, 0) == 0))
        {
            if ((edit_indx > 0 && argvflags.out_filename) ||
                argvflags.out_fileappend)
            {
                edit_flags |= MPGEDIT_FLAGS_APPEND;
            }

            if (!argvflags.s_flag && argvflags.e_flag &&
                 !(argvflags.X_val == 2 &&
                   strcmp(mpgedit_editspec_get_file(next_editspec, 0),
                          XING_HEADER_FILE) == 0))
            {
                printf("Writing edits to file '%s'\n", edit_outfilename);
            }
            edit_ctx = mpgedit_edit_files_init(next_editspec,
                                               edit_outfilename,
                                               edit_flags,
                                               &rsts);
            editing = 1;
            while (rsts == 0 && editing) {
                editing = mpgedit_edit_files(edit_ctx, &rsts);
                if (rsts == 0) {
                    mp3header_print_verbose(
                        edit_ctx,
                        argvflags.s_flag ? -1 : argvflags.v_flag,
                                                &once,
                        &tprev);
                }
            }
            if (rsts == 0) {
                if ((!argvflags.s_flag && 
                      !(argvflags.X_val == 2 && 
                        strcmp(mpgedit_editspec_get_file(
                               next_editspec, 0),
                               XING_HEADER_FILE) == 0)) || 
                    (argvflags.s_flag && !argvflags.e_flag))
                {
                    stats_ptr = mpgedit_edit_stats_ctx(edit_ctx);
                    mpeg_file_stats2str(stats_ptr,
                              mpgedit_edit_total_length_time(edit_ctx), str);
                    printf(str);
                }
            }
            else {
                printf("mpgedit failed: status = %d\n", rsts);
            }
            mpgedit_edit_files_free(edit_ctx);
        }
        if (next_editspec) {
            free(next_editspec);
        }
        if (!argvflags.out_filename && edit_outfilename) {
            free(edit_outfilename);
            edit_outfilename = NULL;
        }
        edit_indx += next_editspec_len;
    } while (edit_indx < edits_len);
    if (edit_outfilename) {
        free(edit_outfilename);
        edit_outfilename = NULL;
    }
    if (argvflags.out_filename) {
        free(argvflags.out_filename);
    }
    if (edit_filename) {
        free(edit_filename);
    }
    
    mpgedit_editspec_free(argvflags.edits);

    if (editfp) {
        fclose(editfp);
        editfp = NULL;
    }
    free(str);

    return 0;
}
